library nueli.core.config;

import 'dart:math';
import 'dart:ui';

import 'package:flutter/cupertino.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

import 'bucket.dart';
import 'data.dart';
import 'util.dart' show Add, AddList, AdvanceList, Subtract, SubtractList;

part 'config/decor.dart';
part 'config/inkwell.dart';
part 'config/layout.dart';
part 'config/photo.dart';
part 'config/radii.dart';
part 'config/tranx.dart';
part 'config/txt.dart';
part 'config/umbra.dart';
part 'config/x3d.dart';

/// ## [Config] ~ Widget Configuration Class.
/// Extend this class to create your own custom
/// configuration for widgets.
abstract class Config<T extends Config<T, IN, OUT>, IN extends ConfigIN<IN, OUT>, OUT extends ConfigOUT<OUT, IN>>
    extends StatefulWidget {
  /// the reference of the instance of [IN]
  /// associated with this [Config].
  /// see: *[_fallback]*
  final Ref<IN> _iN = $();

  /// the fallback or default settings to be used
  /// by this [Config] during [State.build]
  final IN _fallback;

  @protected
  Config({Key? key, required IN fallback})
      : _fallback = fallback,
        super(key: key);

  T createNew();

  /// *@fixed* ~ set [child]
  T child(Widget child) => resolve(child, #child);

  @mustCallSuper
  T resolve<Type>(data, Symbol tag) {
    return update((d) => d.resolve(data, tag));
  }

  @mustCallSuper
  T copy<Type>(T t) => update((d) => d.copy(t().iN));

  @mustCallSuper
  T update(IN Function(IN data) change) {
    return this as T.._iN.set(change(_iN.get ?? _fallback));
  }

  T operator *(double operand) => update((data) => data * operand);

  T operator +(T operand) => update((data) => data + operand().iN);

  T operator -(T operand) => update((data) => data + operand().iN);

  /// subclass should override this method to
  /// define the associated [ConfigOUT]. of the
  /// Call this method to get an instance of the
  /// associated [ConfigOUT] data.
  @mustCallSuper
  OUT call() => (_iN.get ?? _fallback)();

  @override
  ConfigState<T, IN, OUT> createState() => ConfigState();
}

class ConfigState<T extends Config<T, IN, OUT>, IN extends ConfigIN<IN, OUT>, OUT extends ConfigOUT<OUT, IN>>
    extends State<T> {
  @override
  Widget build(BuildContext context) => widget().build(context);
}

abstract class ConfigIN<T extends ConfigIN<T, OUT>, OUT extends ConfigOUT<OUT, T>> {
  Widget? child;

  T createNew();

  OUT call();

  @mustCallSuper
  T operator *(double operand) => createNew().copy(this as T);

  @mustCallSuper
  T operator +(T operand) => createNew().copy(this as T);

  @mustCallSuper
  T operator -(T operand) => createNew().copy(this as T);

  @mustCallSuper
  T resolve(data, Symbol tag) {
    if (data is Widget && tag == #child) child = data;
    return this as T;
  }

  @mustCallSuper
  T copy(T iN) {
    child = iN.child;
    return this as T;
  }
}

abstract class ConfigOUT<T extends ConfigOUT<T, IN>, IN extends ConfigIN<IN, T>> {
  final IN iN;

  ConfigOUT(this.iN);

  @mustCallSuper
  Widget get child => iN.child ?? const SizedBox();

  @mustCallSuper
  Widget inside(Widget child) => child;

  @mustCallSuper
  Widget outside(Widget child) => inside(child);

  @mustCallSuper
  Widget build(BuildContext context) => outside(child);
}

mixin ComplexConfig<T extends ComplexConfig<T, IN, OUT>, IN extends ComplexConfigIN<IN, OUT>,
    OUT extends ComplexConfigOUT<OUT, IN>> on Config<T, IN, OUT> {
  /// *@fixed* ~ set [axis]
  T axis(Axis? axis) => resolve(axis, #axis);

  /// *@fixed* ~ set [children]
  T children(List<Widget>? children) => resolve(children, #children);
}

mixin ComplexConfigIN<T extends ComplexConfigIN<T, OUT>, OUT extends ComplexConfigOUT<OUT, T>> on ConfigIN<T, OUT> {
  Axis? axis;
  List<Widget>? children;

  @override
  T resolve(data, tag) {
    if (data is Axis && tag == #axis) axis = data;
    if (data is List<Widget> && tag == #children) children = data;
    return super.resolve(data, tag);
  }

  @override
  T copy(iN) {
    children = iN.children;
    axis = iN.axis;
    return super.copy(iN);
  }
}

mixin ComplexConfigOUT<T extends ComplexConfigOUT<T, IN>, IN extends ComplexConfigIN<IN, T>> on ConfigOUT<T, IN> {
  @mustCallSuper
  Axis? get axis => iN.axis;

  @mustCallSuper
  List<Widget> get children => iN.children ?? [];

  @mustCallSuper
  Widget combine(List<Widget> children);

  @override
  Widget inside(child) => super.inside(combine([child, ...children]));
}
