part of nueli.core.config;

mixin DecorConfig<T extends DecorConfig<T, IN, OUT>, IN extends DecorConfigIN<IN, OUT>, OUT extends DecorConfigOUT<OUT, IN>>
    on LayoutConfig<T, IN, OUT>, InkwellConfig<T, IN, OUT>, UmbraConfig<T, IN, OUT> {
  /// *@varies* ~ set fill decor paint list
  T gradientFill(List<DecorPaint> hue) => resolve(hue, #gradientFill);

  /// *@varies* ~ set single color
  T bucketFill(Color color) => resolve(DecorPaint(color), #bucketFill);

  /// *@varies* ~ adds color with stop
  T paint(Color color, double stop) => gradientFill(this().paints..add(DecorPaint(color, stop)));

  /// *@varies* ~ set stroke color
  T stroke(double? size, [Color? color]) => resolve(size, #stroke.size).resolve(color, #stroke.color);

  /// *@varies, customized* ~ set gradient fill origin,
  /// ```
  /// if [gradientStyle] is [GradientStyle.linear]
  ///   **[origin] works as [LinearGradient.begin]**
  /// else if [gradientStyle] is [GradientStyle.radial],
  ///   **[origin] works as [RadialGradient.center]**
  /// else if [gradientStyle] is [GradientStyle.sweep],
  ///   **[origin] works as [RadialGradient.center]**.
  T gradientOrigin(Alignment origin) => resolve(origin, #gradientOrigin);

  /// *@varies, customized* ~ set gradient fill origin
  /// ```
  /// if [gradientStyle] is [GradientStyle.linear],
  ///   **[focal] works as [LinearGradient.end]**
  /// else if [gradientStyle] is [GradientStyle.radial],
  ///   **[focal] is negleted**
  /// else if [gradientStyle] is [GradientStyle.sweep],
  ///   **[focal] works as [RadialGradient.focal]**.
  T gradientFocal(Alignment focal) => resolve(focal, #gradientFocal);

  /// *@varies, customized* ~ set decor style
  T gradientStyle(GradientStyle? style) => resolve(style, #gradientStyle);
}

mixin DecorConfigIN<T extends DecorConfigIN<T, OUT>, OUT extends DecorConfigOUT<OUT, T>>
    on LayoutConfigIN<T, OUT>, InkwellConfigIN<T, OUT>, UmbraConfigIN<T, OUT> {
  /// paints
  List<DecorPaint>? paints;
  Alignment? gradientOrigin, gradientFocal;
  Color? stroke;
  double? strokeSize;

  GradientStyle? gradientStyle;

  @override
  T operator *(double operand) {
    return (super * operand)
      ..paints = this().paints.map((e) => e * operand).toList()
      ..gradientOrigin = this().origin * operand
      ..gradientFocal = this().focal * operand
      ..stroke = this().stroke * operand;
  }

  @override
  T operator -(T operand) {
    return (super - operand)
      ..paints = this().paints.minus(operand().paints).toList()
      ..gradientOrigin = this().origin - operand().origin
      ..gradientFocal = this().focal - operand().focal
      ..stroke = this().stroke - operand().stroke
      ..strokeSize = this().strokeSize - operand().strokeSize;
  }

  @override
  T operator +(T operand) {
    return (super + operand)
      ..paints = this().paints.plus(operand().paints).toList()
      ..gradientOrigin = this().origin + operand().origin
      ..gradientFocal = this().focal + operand().focal
      ..stroke = this().stroke + operand().stroke
      ..strokeSize = this().strokeSize + operand().strokeSize;
  }

  @override
  T resolve(data, tag) {
    if (data is DecorPaint && tag == #bucketFill) paints = [data];
    if (data is List<DecorPaint> && tag == #gradientFill) paints = data;
    if (data is Alignment && tag == #gradientOrigin) gradientOrigin = data;
    if (data is Alignment && tag == #gradientFocal) gradientFocal = data;
    if (data is Color && tag == #stroke.color) stroke = data;
    if (data is double && tag == #stroke.size) strokeSize = data;

    if (data is GradientStyle && tag == #gradientStyle) gradientStyle = data;
    return super.resolve(data, tag);
  }

  @override
  T copy(iN) {
    paints = iN.paints ?? paints;
    gradientOrigin = iN.gradientOrigin ?? gradientOrigin;
    gradientFocal = iN.gradientFocal ?? gradientFocal;
    stroke = iN.stroke ?? stroke;
    strokeSize = iN.strokeSize ?? strokeSize;

    gradientStyle = iN.gradientStyle ?? gradientStyle;
    return super.copy(iN);
  }
}

mixin DecorConfigOUT<T extends DecorConfigOUT<T, IN>, IN extends DecorConfigIN<IN, T>>
    on LayoutConfigOUT<T, IN>, UmbraConfigOUT<T, IN>, InkwellConfigOUT<T, IN> {
  List<DecorPaint> get paints => iN.paints ?? [];

  Alignment get origin => iN.gradientOrigin ?? Alignment.centerLeft;

  Alignment get focal => iN.gradientFocal ?? origin * -1;

  Color get stroke => iN.stroke ?? palette.none;

  double get strokeSize => iN.strokeSize ?? 0;

  GradientStyle get style => iN.gradientStyle ?? GradientStyle.linear;

  Gradient? get gradient {
    if (paints.isEmpty) return null;
    if (paints.isUnitary) return null;

    if (style == GradientStyle.linear) {
      return LinearGradient(
        colors: paints,
        begin: origin,
        end: focal,
        stops: paints.stops,
      );
    }
    if (style == GradientStyle.radial) {
      return RadialGradient(
        colors: paints,
        center: origin,
        focal: focal,
        stops: paints.stops,
      );
    }
    if (style == GradientStyle.sweep) {
      return SweepGradient(
        colors: paints,
        center: origin,
        stops: paints.stops,
      );
    }
    return null;
  }

  BoxDecoration get decoration {
    return BoxDecoration(
      color: paints.isEmpty ? palette.none : paints.first,
      gradient: gradient,
      borderRadius: borderRadius,
      border: (strokeSize != 0) ? Border.all(color: stroke, width: strokeSize) : null,
    );
  }

  @override
  Widget inside(child) {
    return Container(
      child: super.inside(child),
      decoration: decoration,
      clipBehavior: clip,
    );
  }
}

class DecorPaint extends Color with Add<DecorPaint>, Subtract<DecorPaint> {
  final double stop;

  DecorPaint(Color color, [this.stop = 1]) : super(color.value);
  DecorPaint.int(int value, [this.stop = 1]) : super(value);

  DecorPaint operator *(double factor) {
    return DecorPaint.int((Color(value) * factor).value, stop * factor);
  }

  @override
  DecorPaint operator -(DecorPaint operand) {
    return DecorPaint.int(
      (Color(value) - Color(operand.value)).value,
      stop - operand.stop,
    );
  }

  @override
  DecorPaint operator +(DecorPaint operand) {
    return DecorPaint.int(
      (Color(value) + Color(operand.value)).value,
      stop + operand.stop,
    );
  }
}

enum GradientStyle { linear, radial, sweep }

extension DecorConfigColor on Color {
  Color operator *(double factor) =>
      Color.fromRGBO((red * factor).round(), (green * factor).round(), (blue * factor).round(), opacity * factor);
  Color operator -(Color color) =>
      Color.fromRGBO(red - color.red, green - color.green, blue - color.blue, opacity - color.opacity);
  Color operator +(Color color) =>
      Color.fromRGBO(red + color.red, green + color.green, blue + color.blue, opacity + color.opacity);
}

extension on List<DecorPaint> {
  List<double> get stops {
    List<double> stops = [];
    for (DecorPaint paint in this) {
      stops.add(paint.stop);
    }
    return stops;
  }
}
