part of nueli.core.config;

mixin LayoutConfig<T extends LayoutConfig<T, IN, OUT>, IN extends LayoutConfigIN<IN, OUT>,
    OUT extends LayoutConfigOUT<OUT, IN>> on ComplexConfig<T, IN, OUT> {
  /// *@varies* ~ set width
  T width(double width) => resolve(width, #width);

  /// *@varies* ~ set height
  T height(double height) => resolve(height, #height);

  /// *@varies* ~ set size
  T size(double x, [double? y]) => width(x).height(y ?? x);

  /// *@varies* ~ set width scale factor
  T widthFactor(double factor) => resolve(factor, #widthFactor);

  /// *@varies* ~ set height scale factor
  T heightFactor(double factor) => resolve(factor, #heightFactor);

  /// *@varies* ~ set padding
  T padding(EdgeInsets? padding) => resolve(padding, #padding);

  /// *@varies* ~ set margin
  T margin(EdgeInsets? margin) => resolve(margin, #margin);

  /// *@varies* ~ set alignment
  T alignment(Alignment? alignment) => resolve(alignment, #alignment);

  /// *@varies* ~ set box constraints
  T constraints(BoxConstraints? constraints) => resolve(constraints, #constraints);

  /// *@shortcut* ~ expand size to maximum
  T get sizeToInfinity => size(double.infinity);

  /// *@shortcut* ~ expand width to maximum
  T get widthToInfinity => width(double.infinity);

  /// *@shortcut* ~ expand height to maximum
  T get heightToInfinity => height(double.infinity);

  /// *@shortcut* ~ align to top left corner
  T get alignmentToStart => alignment(Alignment.topLeft);

  /// *@shortcut* ~ align to center
  T get alignmentToCenter => alignment(Alignment.center);

  /// *@shortcut* ~ align to bottom right corner
  T get alignmentToEnd => alignment(Alignment.bottomRight);

  /// *@config reference* ~ copy umbra
  T copyLayout(LayoutConfig layout) {
    final iN = layout().iN as LayoutConfigIN;
    return update(
      (data) => data
        ..resolve(iN.width, #width)
        ..resolve(iN.height, #height)
        ..resolve(iN.widthFactor, #widthFactor)
        ..resolve(iN.heightFactor, #heightFactor)
        ..resolve(iN.alignment, #alignment)
        ..resolve(iN.padding, #padding)
        ..resolve(iN.margin, #margin)
        ..resolve(iN.constraints, #constraints),
    );
  }
}

mixin LayoutConfigIN<T extends LayoutConfigIN<T, OUT>, OUT extends LayoutConfigOUT<OUT, T>> on ComplexConfigIN<T, OUT> {
  double? width, height, widthFactor, heightFactor;
  EdgeInsets? padding, margin;
  Alignment? alignment;

  BoxConstraints? constraints;

  @override
  @mustCallSuper
  T operator *(double operand) {
    return (super * operand)
      ..width = this().width != null ? this().width! * operand : null
      ..height = this().height != null ? this().height! * operand : null
      ..widthFactor = this().widthFactor * operand
      ..heightFactor = this().heightFactor * operand
      ..padding = this().padding * operand
      ..margin = this().margin * operand
      ..alignment = this().alignment != null ? this().alignment! * operand : null;
  }

  @override
  @mustCallSuper
  T operator -(T operand) {
    return (super - operand)
      ..width = this().width != null ? this().width! - (operand().width ?? this().width!) : null
      ..height = this().height != null ? this().height! - (operand().height ?? this().height!) : null
      ..widthFactor = this().widthFactor - operand().widthFactor
      ..heightFactor = this().heightFactor - operand().heightFactor
      ..padding = this().padding - operand().padding
      ..margin = this().margin - operand().margin
      ..alignment = this().alignment != null ? this().alignment! - (operand().alignment ?? this().alignment!) : null;
  }

  @override
  @mustCallSuper
  T operator +(T operand) {
    return (super + operand)
      ..width = this().width != null ? this().width! + (operand().width ?? this().width!) : null
      ..height = this().height != null ? this().height! + (operand().height ?? this().height!) : null
      ..widthFactor = this().widthFactor + operand().widthFactor
      ..heightFactor = this().heightFactor + operand().heightFactor
      ..padding = this().padding + operand().padding
      ..margin = this().margin + operand().margin
      ..alignment = this().alignment != null ? this().alignment! + (operand().alignment ?? this().alignment!) : null;
  }

  @override
  T resolve(data, tag) {
    if (data is double && tag == #width) width = data;
    if (data is double && tag == #height) height = data;
    if (data is double && tag == #widthFactor) widthFactor = data;
    if (data is double && tag == #heightFactor) heightFactor = data;
    if (data is EdgeInsets && tag == #margin) margin = data;
    if (data is EdgeInsets && tag == #padding) padding = data;
    if (data is Alignment && tag == #alignment) alignment = data;

    if (data is BoxConstraints && tag == #constraints) constraints = data;
    return super.resolve(data, tag);
  }

  @override
  T copy(iN) {
    width = iN.width ?? width;
    height = iN.height ?? height;
    widthFactor = iN.widthFactor ?? widthFactor;
    heightFactor = iN.heightFactor ?? heightFactor;
    margin = iN.margin ?? margin;
    padding = iN.padding ?? padding;
    alignment = iN.alignment ?? alignment;

    constraints = iN.constraints ?? constraints;
    return super.copy(iN);
  }

  @override
  OUT call();
}

mixin LayoutConfigOUT<T extends LayoutConfigOUT<T, IN>, IN extends LayoutConfigIN<IN, T>> on ComplexConfigOUT<T, IN> {
  final Ref<Symbol> _iD = $();

  Size? measuredSize;

  Symbol get iD {
    _iD.set(_iD.get ?? ID.newInScope);
    return _iD.get!;
  }

  @mustCallSuper
  double? get width => iN.width;

  @mustCallSuper
  double? get height => iN.height;

  @mustCallSuper
  double get widthFactor => iN.widthFactor ?? 1;

  @mustCallSuper
  double get heightFactor => iN.heightFactor ?? 1;

  @mustCallSuper
  EdgeInsets get padding => iN.padding ?? EdgeInsets.zero;

  @mustCallSuper
  EdgeInsets get margin => iN.margin ?? EdgeInsets.zero;

  @mustCallSuper
  Alignment? get alignment => iN.alignment;

  @mustCallSuper
  double? get shortestSide {
    if (width == null) return null;
    if (height == null) return null;
    return min(width!, height!);
  }

  @mustCallSuper
  double get longestSide => max(width ?? 0, height ?? 0);

  @mustCallSuper
  BoxConstraints? get constraints => iN.constraints;

  @mustCallSuper
  LayoutData get data {
    return LayoutData(
      width: width,
      height: height,
      axis: axis ?? Axis.horizontal,
      alignment: alignment ?? Alignment.topLeft,
      isComplex: children.length > 1,
      spread: false,
      stretch: false,
    );
  }

  @mustCallSuper
  LayoutData get parentData => LayoutData.bank[iD] ?? LayoutData();

  @override
  Widget combine(children) {
    if (children.isUnitary) {
      final child = Container(
        child: children.first,
        width: parentData.hasExpandedWidth ? width : (data.hasExpandedWidth ? null : width),
        height: parentData.hasExpandedHeight ? height : (data.hasExpandedHeight ? null : height),
        padding: padding,
        alignment: parentData.isComplex
            ? parentData.hasExpandedMain
                ? data.hasExpandedAxis(parentData.axis)
                    ? alignment
                    : null
                : null
            : alignment,
        constraints: constraints,
      );

      return parentData.isComplex
          ? parentData.hasExpandedMain
              ? data.hasExpandedAxis(parentData.axis)
                  ? Expanded(child: child)
                  : Align(child: child, alignment: alignment ?? Alignment.center)
              : Align(child: child, alignment: alignment ?? Alignment.center)
          : child;
    }

    if (children.isNotEmpty) {
      final child = Container(
          padding: padding,
          child: data.axisIsVertical
              ? Column(
                  children: children,
                  mainAxisAlignment: data.mainAxisAlignment,
                  crossAxisAlignment: data.crossAxisAlignment,
                  mainAxisSize: parentData.hasExpandedAxis(data.axis) ? data.mainAxisSize : MainAxisSize.min,
                )
              : Row(
                  children: children,
                  mainAxisAlignment: data.mainAxisAlignment,
                  crossAxisAlignment: data.crossAxisAlignment,
                  mainAxisSize: parentData.hasExpandedAxis(data.axis) ? data.mainAxisSize : MainAxisSize.min,
                ));
      return parentData.isComplex
          ? parentData.hasExpandedMain
              ? data.hasExpandedAxis(parentData.axis)
                  ? Expanded(child: child)
                  : child
              : child
          : child;
    }
    return child;
  }

  @override
  Widget outside(Widget child) => Padding(padding: margin, child: super.outside(child));
}

class LayoutWidget extends SingleChildRenderObjectWidget {
  final Function(Size size) onChange;

  const LayoutWidget({
    Key? key,
    required this.onChange,
    required Widget child,
  }) : super(key: key, child: child);

  @override
  RenderObject createRenderObject(BuildContext context) {
    return LayoutRenderObject(onChange);
  }
}

class LayoutRenderObject extends RenderProxyBox {
  Size? oldSize;
  final Function(Size size) onChange;

  LayoutRenderObject(this.onChange);

  @override
  void performLayout() {
    super.performLayout();

    Size newSize = child!.size;
    if (oldSize == newSize) return;

    oldSize = newSize;
    WidgetsBinding.instance!.addPostFrameCallback((_) {
      onChange(newSize);
    });
  }
}

class LayoutData {
  static Map<Symbol, LayoutData> bank = {};

  final double? height;
  final double? width;
  final bool isComplex;
  final bool spread;
  final bool stretch;
  final Alignment alignment;
  final Axis axis;

  LayoutData({
    this.height,
    this.width,
    this.isComplex = false,
    this.spread = false,
    this.stretch = false,
    this.alignment = Alignment.topLeft,
    this.axis = Axis.horizontal,
  });

  double? sizeOfAxis(Axis axis) => axis == Axis.horizontal ? width : height;

  bool hasExpandedAxis(Axis axis) => sizeOfAxis(axis) == double.infinity;

  bool get axisIsVertical => axis == Axis.vertical;
  bool get axisIsHorizontal => axis == Axis.horizontal;

  bool get hasExpandedHeight => hasExpandedAxis(Axis.vertical);
  bool get hasExpandedWidth => hasExpandedAxis(Axis.horizontal);

  bool get hasExpandedMain => axisIsVertical ? hasExpandedHeight : hasExpandedWidth;
  bool get hasExpandedCross => axisIsVertical ? hasExpandedHeight : hasExpandedWidth;

  MainAxisSize get mainAxisSize {
    if (spread) return MainAxisSize.max;
    if (axisIsVertical) return hasExpandedHeight ? MainAxisSize.max : MainAxisSize.min;
    if (axisIsHorizontal) return hasExpandedWidth ? MainAxisSize.max : MainAxisSize.min;
    return MainAxisSize.min;
  }

  MainAxisAlignment get mainAxisAlignment {
    if (spread) return MainAxisAlignment.spaceBetween;
    if (axisIsVertical) {
      if (alignment.y == -1) return MainAxisAlignment.start;
      if (alignment.y == 0) return MainAxisAlignment.center;
      if (alignment.y == 1) return MainAxisAlignment.end;
    }
    if (axisIsHorizontal) {
      if (alignment.x == -1) return MainAxisAlignment.start;
      if (alignment.x == 0) return MainAxisAlignment.center;
      if (alignment.x == 1) return MainAxisAlignment.end;
    }
    return MainAxisAlignment.center;
  }

  CrossAxisAlignment get crossAxisAlignment {
    if (stretch) return CrossAxisAlignment.stretch;
    if (axisIsVertical) {
      if (alignment.y == -1) return CrossAxisAlignment.start;
      if (alignment.y == 0) return CrossAxisAlignment.center;
      if (alignment.y == 1) return CrossAxisAlignment.end;
    }
    if (axisIsHorizontal) {
      if (alignment.x == -1) return CrossAxisAlignment.start;
      if (alignment.x == 0) return CrossAxisAlignment.center;
      if (alignment.x == 1) return CrossAxisAlignment.end;
    }
    return CrossAxisAlignment.center;
  }
}
