part of nueli.core.config;

/// Tranx shortcut
Tranx get $tranx => Tranx();

/// Tranx ~ [AnimatedWidget] and [Tween] configuration class
/// All commands are fixed and instantaneous.
/// ```
class Tranx extends Config<Tranx, TranxIN, TranxOUT> {
  Tranx({Key? key}) : super(key: key, fallback: TranxIN());

  Tranx animate({
    Tween<double>? scale,
    Tween<Offset>? translate,
    Tween<double>? height,
    Tween<double>? width,
    Tween<double>? rotation,
    Tween<double>? opacity,
    Tween<Config>? configTween,
  }) {
    if (scale != null) resolve(_Tween.tween(_TweenType.scale, scale), #tween);
    if (translate != null) resolve(_Tween.tween(_TweenType.translate, translate), #tween);
    if (height != null) resolve(_Tween.tween(_TweenType.height, height), #tween);
    if (width != null) resolve(_Tween.tween(_TweenType.width, width), #tween);
    if (rotation != null) resolve(_Tween.tween(_TweenType.rotation, rotation), #tween);
    if (opacity != null) resolve(_Tween.tween(_TweenType.opacity, opacity), #tween);
    if (configTween != null) resolve(configTween, #tween);
    return this;
  }

  Tranx height(double begin, double end, [Alignment alignment = Alignment.center]) {
    return animate(width: Tween(begin: begin, end: end));
  }

  Tranx width(double begin, double end, [Alignment alignment = Alignment.center]) {
    return animate(height: Tween(begin: begin, end: end));
  }

  Tranx size(Size begin, Size end, [Alignment alignment = Alignment.center]) {
    return height(begin.width, end.width).width(begin.height, end.height);
  }

  Tranx scale(double begin, double end, [Alignment alignment = Alignment.center]) {
    return animate(scale: Tween(begin: begin, end: end));
  }

  Tranx translate(Offset begin, Offset end) => animate(translate: Tween(begin: begin, end: end));

  Tranx rotate(double begin, double end) => animate(rotation: Tween(begin: begin, end: end));

  Tranx fade(double begin, double end) => animate(opacity: Tween(begin: begin, end: end));

  Tranx decor(DecorConfig begin, DecorConfig end) => animate(configTween: Tween<DecorConfig>(begin: begin, end: end));

  Tranx ink(InkwellConfig begin, InkwellConfig end) => animate(configTween: Tween<InkwellConfig>(begin: begin, end: end));

  Tranx media(PhotoConfig begin, PhotoConfig end) => animate(configTween: Tween<PhotoConfig>(begin: begin, end: end));

  Tranx radii(RadiiConfig begin, RadiiConfig end) => animate(configTween: Tween<RadiiConfig>(begin: begin, end: end));

  Tranx umbra(UmbraConfig begin, UmbraConfig end) => animate(configTween: Tween<UmbraConfig>(begin: begin, end: end));

  Tranx txt(TxtConfig begin, TxtConfig end) => animate(configTween: Tween<TxtConfig>(begin: begin, end: end));

  Tranx x3D(X3DConfig begin, X3DConfig end) => animate(configTween: Tween<X3DConfig>(begin: begin, end: end));

  Tranx animation(Animation<double>? animation) => resolve(animation, #animation);

  @override
  Tranx createNew() => Tranx();
}

class TranxIN extends ConfigIN<TranxIN, TranxOUT> {
  List<Tween>? tween;
  Animation<double>? animation;
  Curve? curve;
  Curve? reverseCurve;
  Duration? duration;
  Duration? reverseDuration;

  @override
  TranxIN resolve(data, tag) {
    if (data is Tween && tag == #tween) tween = this().tween..add(data);
    if (data is Animation<double> && tag == #animation) animation = data;
    if (data is Curve && tag == #curve) curve = data;
    if (data is Curve && tag == #reverseCurve) reverseCurve = data;
    if (data is Duration && tag == #duration) duration = data;
    if (data is Duration && tag == #reverseDuration) reverseDuration = data;
    return super.resolve(data, tag);
  }

  @override
  TranxIN copy(iN) {
    tween = iN.tween ?? tween;
    animation = iN.animation ?? animation;
    curve = iN.curve ?? curve;
    reverseCurve = iN.reverseCurve ?? reverseCurve;
    duration = iN.duration ?? duration;
    reverseDuration = iN.reverseDuration ?? reverseDuration;
    return super.copy(iN);
  }

  @override
  TranxOUT call() => TranxOUT(this);

  @override
  TranxIN createNew() => TranxIN();
}

class TranxOUT extends ConfigOUT<TranxOUT, TranxIN> implements Animation<double> {
  TranxOUT(TranxIN iN) : super(iN);

  List<Tween> get tween => iN.tween ?? [];

  Animation<double>? get animation => iN.animation;

  Curve get curve => iN.curve ?? Curves.easeOutExpo;

  Curve get reverseCurve => iN.reverseCurve ?? curve.flipped;

  Duration get duration => iN.duration ?? const Duration(milliseconds: 500);

  Duration get reverseDuration => iN.reverseDuration ?? duration;

  @override
  Widget inside(child) {
    if (animation == null) return super.inside(child);
    Animation<double> anim = CurvedAnimation(parent: animation!, curve: curve);
    for (Tween tween in tween) {
      if (tween is _Tween<double>) {
        if (tween.type == _TweenType.scale) {
          child = ScaleTransition(scale: tween.animate(anim), child: child);
        }
        if (tween.type == _TweenType.height) {
          child = SizeTransition(axis: Axis.vertical, sizeFactor: tween.animate(anim), child: child);
        }
        if (tween.type == _TweenType.width) {
          child = SizeTransition(axis: Axis.horizontal, sizeFactor: tween.animate(anim), child: child);
        }
        if (tween.type == _TweenType.opacity) {
          child = FadeTransition(opacity: tween.animate(anim), child: child);
        }
        if (tween.type == _TweenType.rotation) {
          child = RotationTransition(turns: tween.animate(anim), child: child);
        }
      }
      if (tween is _Tween<Offset>) {
        if (tween.type == _TweenType.translate) {
          child = SlideTransition(position: tween.animate(anim), child: child);
        }
      }
      if (tween is Tween<Config>) child = ConfigTransition(animation: tween.animate(anim), child: child);
    }
    return child;
  }

  @override
  void addListener(VoidCallback listener) => animation!.addListener(listener);

  @override
  void addStatusListener(AnimationStatusListener listener) => animation!.addStatusListener(listener);

  @override
  Animation<U> drive<U>(Animatable<U> child) => animation!.drive(child);

  @override
  bool get isCompleted => animation!.isCompleted;

  @override
  bool get isDismissed => animation!.isDismissed;

  @override
  void removeListener(VoidCallback listener) => animation!.removeListener(listener);

  @override
  void removeStatusListener(AnimationStatusListener listener) => animation!.removeStatusListener(listener);

  @override
  AnimationStatus get status => animation!.status;

  @override
  String toStringDetails() => animation!.toStringDetails();

  @override
  double get value => animation!.value;
}

class ConfigTransition extends AnimatedWidget {
  final Widget child;

  const ConfigTransition({
    Key? key,
    required Animation<Config> animation,
    this.child = const SizedBox(),
  }) : super(key: key, listenable: animation);

  Animation<Config> get animation => listenable as Animation<Config>;

  @override
  Widget build(BuildContext context) {
    return animation.value.child(child);
  }
}

enum _TweenType { scale, translate, height, width, rotation, opacity }

class _Tween<T> extends Tween<T> {
  final _TweenType type;
  _Tween(this.type, {T? begin, T? end}) : super(begin: begin, end: end);
  _Tween.tween(this.type, Tween<T> tween) : super(begin: tween.begin, end: tween.end);
}
