part of nueli.core.config;

/// ## [UmbraConfig] ~ [BoxShadow] configuration class.
/// Gives you access to configure a [BoxShadow].
mixin UmbraConfig<T extends UmbraConfig<T, IN, OUT>, IN extends UmbraConfigIN<IN, OUT>, OUT extends UmbraConfigOUT<OUT, IN>>
    on Config<T, IN, OUT> {
  /// *@varies* ~ shadow color
  T shadowColor(Color color, [int? alpha]) {
    return resolve(color.withAlpha(alpha ?? palette.shade.alpha), #shadowColor);
  }

  /// *@varies* ~ set shadow offset ([x], [y])
  T shadowOffset(double x, [double? y]) {
    return resolve(Offset(x, y ?? x), #shadowOffset);
  }

  /// *@varies* ~ set shadow spread radius
  T shadowSpread(double spread) => resolve(spread, #shadowSpread);

  /// *@varies* ~ set shadow blur radius
  T shadowBlur(double blur) => resolve(blur, #shadowBlur);

  /// *@fixed* ~ set shadow layer position
  T shadowPosition(DecorationPosition? position) => resolve(position, #shadowPosition);

  /// *@config reference* ~ copy umbra
  T copyUmbra(UmbraConfig umbra) {
    final iN = umbra().iN as UmbraConfigIN;
    return update(
      (data) => data
        ..resolve(iN.shadowBlur, #shadowBlur)
        ..resolve(iN.shadowColor, #shadowColor)
        ..resolve(iN.shadowOffset, #shadowOffset)
        ..resolve(iN.shadowPosition, #shadowPosition)
        ..resolve(iN.shadowSpread, #shadowSpread),
    );
  }
}

mixin UmbraConfigIN<T extends UmbraConfigIN<T, OUT>, OUT extends UmbraConfigOUT<OUT, T>> on ConfigIN<T, OUT> {
  /// *@varies* ~ color of the shadow.
  Color? shadowColor;

  /// *@varies* ~ offset of the shadow.
  Offset? shadowOffset;

  /// *@varies* ~ spread radius of the shadow.
  double? shadowSpread;

  /// *@varies* ~ blur radius of the shadow.
  double? shadowBlur;

  /// *@varies* ~ position of the shadow.
  DecorationPosition? shadowPosition;

  @override
  T operator *(double operand) {
    return (super * operand)
      ..shadowColor = this().shadowColor * operand
      ..shadowOffset = this().shadowOffset * operand
      ..shadowSpread = this().shadowSpread * operand
      ..shadowBlur = this().shadowBlur * operand;
  }

  @override
  T operator -(T operand) {
    return (super - operand)
      ..shadowColor = this().shadowColor - operand().shadowColor
      ..shadowOffset = this().shadowOffset - operand().shadowOffset
      ..shadowSpread = this().shadowSpread - operand().shadowSpread
      ..shadowBlur = this().shadowBlur - operand().shadowBlur;
  }

  @override
  T operator +(T operand) {
    return (super + operand)
      ..shadowColor = this().shadowColor + operand().shadowColor
      ..shadowOffset = this().shadowOffset + operand().shadowOffset
      ..shadowSpread = this().shadowSpread + operand().shadowSpread
      ..shadowBlur = this().shadowBlur + operand().shadowBlur;
  }

  @override
  T resolve(data, tag) {
    if (data is Color && tag == #shadowColor) shadowColor = data;
    if (data is double && tag == #shadowBlur) shadowBlur = data;
    if (data is double && tag == #shadowSpread) shadowSpread = data;
    if (data is Offset && tag == #shadowOffset) shadowOffset = data;

    if (data is DecorationPosition && tag == #position) shadowPosition = data;
    return super.resolve(data, tag);
  }

  @override
  T copy(iN) {
    shadowColor = iN.shadowColor ?? shadowColor;
    shadowBlur = iN.shadowBlur ?? shadowBlur;
    shadowSpread = iN.shadowSpread ?? shadowSpread;
    shadowOffset = iN.shadowOffset ?? shadowOffset;

    shadowPosition = iN.shadowPosition ?? shadowPosition;
    return super.copy(iN);
  }
}

mixin UmbraConfigOUT<T extends UmbraConfigOUT<T, IN>, IN extends UmbraConfigIN<IN, T>> on ConfigOUT<T, IN> {
  Color get shadowColor => iN.shadowColor ?? palette.none;
  Offset get shadowOffset => iN.shadowOffset ?? Offset.zero;
  double get shadowSpread => iN.shadowSpread ?? 0;
  double get shadowBlur => iN.shadowBlur ?? 0;

  DecorationPosition get shadowPosition {
    return iN.shadowPosition ?? DecorationPosition.background;
  }

  BoxShadow get boxShadow {
    return BoxShadow(
      color: shadowColor,
      offset: shadowOffset,
      blurRadius: shadowBlur,
      spreadRadius: shadowSpread,
    );
  }

  List<BoxShadow> get boxShadowDeep => [boxShadow];

  @override
  Widget outside(child) {
    return DecoratedBox(
      child: super.outside(child),
      decoration: BoxDecoration(boxShadow: boxShadowDeep),
      position: shadowPosition,
    );
  }
}
