library nueli.core.data;

import 'dart:async';

import 'package:flutter/cupertino.dart';

/// ## [Ref] Shortcut
Ref<T> $<T>([T? data]) => Ref<T>.create(data);

/// Shortcut to begin an event Algorithm
Algorithm get algog => Algorithm();

/// Data ID Class
abstract class ID {
  static int _scopePrev = -1;

  static Symbol get newInScope {
    _scopePrev++;
    return Symbol('$_scopePrev');
  }
}

/// ## [Ref] ~ Data reference
/// * Having problems with non immutable classes?
/// * have a few fields you will like to edit or
///   change any where in your code?
/// * need raw global access in to a value in
///   your code?
///
/// Well, your problem is solved. You can now
/// reference your [Object] using the [Ref] class
/// and easily set and get the value anytime and
/// anywhere in your code. [Ref] helps to assign
/// a [key] to your Object. Making it available
/// globally in your code.
///
/// ```
/// var data = Ref.of(#data).get;
/// Ref.of(#data).set(data);
/// Ref.of(#data).get = data}
/// Ref.create(data)
/// ```
class Ref<T> implements ChangeNotifier {
  static final Map<Symbol, ValueNotifier> _baseMap = {};
  static Symbol? receipt;

  late final Symbol key;

  /// create a new [Ref] instance pointing to
  /// [key]. The created reference will host [null]
  /// or The value of the old reference if any.
  /// ```
  /// final ref = Ref.create(value);
  /// ```
  Ref.of(this.key) {
    receipt = key;
    _baseMap[key] = ValueNotifier(null);
  }

  /// create a new [Ref] instance pointing to a
  /// unique [key] generated automatically. To
  /// access the generated key, see: [receipt].
  /// The created reference will host [value]
  /// ```
  /// final ref = Ref.create(value);
  /// ```
  factory Ref.create(T? value) => Ref<T>.of(ID.newInScope).set(value);

  T? get get => _baseMap[key]!.value;

  set get(T? value) => _baseMap[key] = ValueNotifier(value);

  Ref<T> set(T? data) {
    notifyListeners();
    return this..get = data;
  }

  @override
  String toString() => get.toString();

  @override
  int get hashCode => get.hashCode;

  @override
  bool operator ==(Object other) => get == other;

  @override
  Type get runtimeType => get.runtimeType;

  @override
  bool get hasListeners => _baseMap[key]!.hasListeners;

  @override
  void addListener(VoidCallback listener) => _baseMap[key]!.addListener(listener);

  @override
  void removeListener(VoidCallback listener) => _baseMap[key]!.removeListener(listener);

  @override
  void dispose() => _baseMap[key]!.dispose();

  @override
  void notifyListeners() => _baseMap[key]!.notifyListeners();
}

/// Data Algorithm
class Algorithm<T> {
  final Ref<List<Future<T> Function()>> _flow = $([]);

  Algorithm([List<Future<T> Function()>? flow]) {
    _flow.get = (flow ?? []).toList();
  }

  Algorithm<T> then(Future<T> Function()? operand) {
    if (operand != null) _flow.get = [...(_flow.get ?? []), operand];
    return this;
  }

  Algorithm<T> concat(Algorithm<T>? operand) {
    if (operand != null) _flow.get = [...(_flow.get ?? []), ...(operand._flow.get ?? [])];
    return this;
  }

  Future<T?> call({bool reverse = false}) async {
    T? result;
    for (Future<T> Function() task in (_flow.get ?? [])) {
      await task().then((value) => result = value);
    }
    return result;
  }
}
